from collections import defaultdict
from enum import Enum
from typing import Dict, Iterable, Optional, Set, Type
from pathlib import Path

from xlsxwriter import Workbook

from .config import CONFIG as _CONFIG
from .entity import Entity, Place, Person, Regest


class EntityType(Enum):
    PERSON = (Person, 'O')
    PLACE = (Place, 'M')

    @classmethod
    def from_entity_class(cls, entity_class: Optional[Type[Entity]]) -> Optional['EntityType']:
        if entity_class is None:
            return None
        for value in cls:
            if value.to_entity_class() == entity_class:
                return value
        raise ValueError('Unexpected entity class: {}'.format(entity_class))

    @classmethod
    def from_symbol(cls, symbol: Optional[str]) -> Optional['EntityType']:
        if symbol is None:
            return None
        for value in cls:
            if str(value) == symbol:
                return value
        raise ValueError('Unexpected symbol: {}'.format(symbol))

    def __str__(self) -> str:
        return self.value[1]

    def to_entity_class(self) -> Type[Entity]:
        return self.value[0]

    @classmethod
    def list_symbols(cls) -> str:
        return ', '.join(str(value) for value in cls)


class Approval(Enum):
    APPROVED = 'A'
    DISAPPROVED = 'N'
    TOO_LONG = '>'
    TOO_SHORT = '<'

    @classmethod
    def from_symbol(cls, symbol: Optional[str]) -> Optional['Approval']:
        if symbol is None:
            return None
        for value in cls:
            if str(value) == symbol:
                return value
        raise ValueError('Unexpected symbol: {}'.format(symbol))

    def __str__(self) -> str:
        return self.value

    @classmethod
    def list_symbols(cls) -> str:
        return ', '.join(str(value) for value in cls)


class Patch:
    def __init__(self, entity: Entity, canonical_form: Optional[str],
                 approval: Optional[Approval], note: Optional[str]):
        self.entity = entity
        self.canonical_form = canonical_form
        self.approval = approval
        self.note = note


class Patchset:
    CONFIG = _CONFIG['patch.Patchset']
    ROOT_PATH = Path(CONFIG['root_path'])
    MINIMUM_LENGTH = CONFIG.getint('minimum_length')
    MAXIMUM_LENGTH = CONFIG.getint('maximum_length')

    @classmethod
    def create_patchset_template(cls, filename: str, entities: Iterable[Entity],
                                 entities_to_regests: Dict[Entity, Set[Regest]]) -> None:

        entities = sorted(entities)

        # Create a plaintext file for power users
        with (cls.ROOT_PATH / filename).with_suffix('.txt').open('wt') as f:
            for entity in entities:
                print(entity, file=f)

        # Create a spreadsheet for standard users
        with Workbook((cls.ROOT_PATH / filename).with_suffix('.xlsx')) as workbook:

            # Declare formats
            default_format = workbook.add_format()

            bold_format = workbook.add_format({'bold': True})

            malformed_format = workbook.add_format({'bg_color': '#f4cccc'})
            well_formed_format = workbook.add_format({'bg_color': '#d9ead3'})

            disapproved_format = workbook.add_format({'bg_color': '#f4cccc'})
            approved_format = workbook.add_format({'bg_color': '#d9ead3'})
            too_short_or_long_format = workbook.add_format({'bg_color': '#fff2cc'})

            canonicalized_format = workbook.add_format({'bg_color': '#d9ead3'})

            # Sort entities by initial letters
            letters = defaultdict(lambda: list())
            for entity in entities:
                letter, *_ = str(entity)
                letter = letter.upper()
                letters[letter].append(entity)

            # Create worksheets for individual initial letters
            for letter, letter_entities in sorted(letters.items()):
                worksheet = workbook.add_worksheet(letter)

                # Set column width
                worksheet.set_column('A:A', 50)
                worksheet.set_column('B:B', 20)
                worksheet.set_column('C:C', 15)
                worksheet.set_column('D:D', 50)
                worksheet.set_column('E:E', 50)

                # Add headers
                worksheet.write('A1', 'Entita', bold_format)
                worksheet.write('B1', f'Použít? ({Approval.list_symbols()})', bold_format)
                worksheet.write_comment('B1', 'Uveďte, jestli si přejete entitu použít v '
                                        'trénovací datové sadě pro rozpoznávání jmenných '
                                        f'entit.\n\n{Approval.APPROVED}: Ano, přeji si entitu použít.\n\n'
                                        f'{Approval.DISAPPROVED}: Ne, přeji si entitu vynechat z datové sady.\n\n'
                                        f'{Approval.TOO_SHORT}: Entita je příliš krátká, pravděpodobně ji budeme '
                                        'chtít vynechat z datové sady, nebo ji budeme chtít '
                                        'přisoudit menší váhu.\n\n'
                                        f'{Approval.TOO_LONG}: Entita je příliš dlouhá, pravděpodobně ji budeme '
                                        'chtít vynechat z datové sady, nebo ji budeme chtít '
                                        'přisoudit menší váhu.')
                worksheet.write('C1', f'Typ entity ({EntityType.list_symbols()})', bold_format)
                worksheet.write_comment('C1', f'Uveďte typ entity:\n\n{EntityType.PERSON}: Osoba\n\n'
                                        f'{EntityType.PLACE}: Místo')
                worksheet.write('D1', 'Kanonická forma', bold_format)
                worksheet.write_comment('D1', 'Uveďte kanonickou formu entity, kterou '
                                        'budeme následně fulltextově hledat v dokumentech. '
                                        'Pokud lze kanonickou formu získat automatizovaně '
                                        '(např. odstraněním interpunkce zleva a zprava), '
                                        'použijte prosím poznámkový aparát.')
                worksheet.write('E1', 'Poznámky', bold_format)
                worksheet.write_comment('E1', 'Volitelně uveďte dodatečné informace ke své '
                                        'anotaci.')

                # List entities
                assert cls.MINIMUM_LENGTH <= cls.MAXIMUM_LENGTH
                for row_number, entity in enumerate(letter_entities):
                    row_number += 1
                    regest = sorted(entities_to_regests[entity])[0]

                    worksheet.write_url(row_number, 0, regest.document_url, string=str(entity))

                    if len(entity) < cls.MINIMUM_LENGTH:
                        worksheet.write(row_number, 1, str(Approval.TOO_SHORT))
                    elif len(entity) > cls.MAXIMUM_LENGTH:
                        worksheet.write(row_number, 1, str(Approval.TOO_LONG))

                # Add an autofilter, so that users can reorder the entities as they need
                worksheet.autofilter(0, 0, len(letter_entities), 4)

                # Set conditional formatting for approval
                blank_conditional_format = {'type': 'blanks', 'stop_if_true': True, 'format': default_format}
                disapproved_conditional_format = {'type': 'cell', 'criteria': '==',
                                                  'value': f'"{Approval.DISAPPROVED}"',
                                                  'format': disapproved_format}
                approved_conditional_format = {'type': 'cell', 'criteria': '==',
                                               'value': f'"{Approval.APPROVED}"',
                                               'format': approved_format}
                too_short_conditional_format = {'type': 'cell', 'criteria': '==',
                                                'value': f'"{Approval.TOO_SHORT}"',
                                                'format': too_short_or_long_format}
                too_long_conditional_format = {'type': 'cell', 'criteria': '==',
                                               'value': f'"{Approval.TOO_LONG}"',
                                               'format': too_short_or_long_format}

                approval_format_range = [1, 1, len(letter_entities), 1]

                worksheet.conditional_format(*approval_format_range, blank_conditional_format)
                worksheet.conditional_format(*approval_format_range, disapproved_conditional_format)
                worksheet.conditional_format(*approval_format_range, approved_conditional_format)
                worksheet.conditional_format(*approval_format_range, too_short_conditional_format)
                worksheet.conditional_format(*approval_format_range, too_long_conditional_format)

                # Set conditional formatting for entity types
                place_conditional_format = {'type': 'cell', 'criteria': '==',
                                            'value': f'"{EntityType.PLACE}"',
                                            'format': well_formed_format}
                person_conditional_format = {'type': 'cell', 'criteria': '==',
                                             'value': f'"{EntityType.PERSON}"',
                                             'format': well_formed_format}
                malformed_conditional_format = {
                    'type': 'formula',
                    'criteria': f'=AND($C2<>"{EntityType.PERSON}", $C2<>"{EntityType.PLACE}")',
                    'stop_if_true': True, 'format': malformed_format,
                }

                correct_entity_type_format_range = [1, 2, len(letter_entities), 2]

                worksheet.conditional_format(*correct_entity_type_format_range, blank_conditional_format)
                worksheet.conditional_format(*correct_entity_type_format_range, place_conditional_format)
                worksheet.conditional_format(*correct_entity_type_format_range, person_conditional_format)
                worksheet.conditional_format(*correct_entity_type_format_range, malformed_conditional_format)

                # Set conditional formatting for canonical form
                canonicalized_conditional_format = {'type': 'no_blanks', 'format': canonicalized_format}

                canonical_form_format_range = [1, 3, len(letter_entities), 3]

                worksheet.conditional_format(*canonical_form_format_range, blank_conditional_format)
                worksheet.conditional_format(*canonical_form_format_range, canonicalized_conditional_format)
